#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.SupportVectorMachine;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Evaluators;
using HeuristicLab.Parameters;
using HeuristicLab.Optimization;
using HeuristicLab.Operators;
using HeuristicLab.Problems.DataAnalysis.Regression.Symbolic;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using System.Collections.Generic;
using HeuristicLab.Problems.DataAnalysis.Regression;
using HeuristicLab.Problems.DataAnalysis.MultiVariate.TimeSeriesPrognosis.Symbolic.Interfaces;
using HeuristicLab.Problems.DataAnalysis.MultiVariate.Evaluators;

namespace HeuristicLab.Problems.DataAnalysis.MultiVariate.TimeSeriesPrognosis.Symbolic.Evaluators {
  [Item("SymbolicTimeSeriesPrognosisScaledNormalizedMseEvaluator", "")]
  [StorableClass]
  public class SymbolicTimeSeriesPrognosisScaledNormalizedMseEvaluator : SymbolicTimeSeriesPrognosisEvaluator {

    public SymbolicTimeSeriesPrognosisScaledNormalizedMseEvaluator()
      : base() {
    }

    public override double Evaluate(SymbolicExpressionTree tree, MultiVariateDataAnalysisProblemData problemData, ISymbolicTimeSeriesExpressionInterpreter interpreter, IEnumerable<int> rows, int predictionHorizon, DoubleArray lowerEstimationLimit, DoubleArray upperEstimationLimit) {
      return Calculate(tree, problemData, interpreter, rows, predictionHorizon, lowerEstimationLimit, upperEstimationLimit);
    }

    public static double Calculate(SymbolicExpressionTree tree, MultiVariateDataAnalysisProblemData problemData, ISymbolicTimeSeriesExpressionInterpreter interpreter, IEnumerable<int> rows, int predictionHorizon, DoubleArray lowerEstimationLimit, DoubleArray upperEstimationLimit) {
      double[] alpha, beta;
      double quality;

      // calculate scaling parameters based on one-step-predictions
      IEnumerable<string> selectedTargetVariables = from item in problemData.TargetVariables
                                                    where problemData.TargetVariables.ItemChecked(item)
                                                    select item.Value;
      int dimension = selectedTargetVariables.Count();

      IEnumerable<double[]> oneStepPredictions =
        interpreter.GetSymbolicExpressionTreeValues(tree, problemData.Dataset, selectedTargetVariables, rows, 1);
      IEnumerable<double[]> originalValues = from row in rows
                                             select (from targetVariable in selectedTargetVariables
                                                     select problemData.Dataset[targetVariable, row]).ToArray();
      alpha = new double[dimension];
      beta = new double[dimension];

      CalculateScalingParameters(originalValues, oneStepPredictions, ref beta, ref alpha);

      // calculate the quality for the full horizon
      quality = CalculateWithScaling(tree, problemData, interpreter,
        rows, predictionHorizon,
        lowerEstimationLimit, upperEstimationLimit,
        beta, alpha);
      return quality;
    }

    public static double CalculateWithScaling(SymbolicExpressionTree tree, MultiVariateDataAnalysisProblemData problemData,
      ISymbolicTimeSeriesExpressionInterpreter interpreter,
      IEnumerable<int> rows, int predictionHorizon,
      DoubleArray lowerEstimationLimit, DoubleArray upperEstimationLimit,
      double[] beta, double[] alpha) {
      IEnumerable<string> selectedTargetVariables = from targetVariable in problemData.TargetVariables
                                                    where problemData.TargetVariables.ItemChecked(targetVariable)
                                                    select targetVariable.Value;

      IEnumerable<double[]> estimatedValues =
        interpreter.GetScaledSymbolicExpressionTreeValues(tree, problemData.Dataset, selectedTargetVariables,
        rows, predictionHorizon, beta, alpha);

      IEnumerable<double[]> originalValues = from row in rows
                                             from step in Enumerable.Range(0, predictionHorizon)
                                             select (from targetVariable in selectedTargetVariables
                                                     select problemData.Dataset[targetVariable, row + step]).ToArray();

      var evaluator = new OnlineMultiVariateEvaluator<OnlineNormalizedMeanSquaredErrorEvaluator>();

      var estimatedValuesEnumerator = estimatedValues.GetEnumerator();
      var originalValuesEnumerator = originalValues.GetEnumerator();
      while (originalValuesEnumerator.MoveNext() & estimatedValuesEnumerator.MoveNext()) {
        double[] original = originalValuesEnumerator.Current;
        double[] estimated = estimatedValuesEnumerator.Current;
        for (int i = 0; i < estimated.Length; i++) {
          if (double.IsNaN(estimated[i])) estimated[i] = upperEstimationLimit[i];
          else estimated[i] = Math.Min(upperEstimationLimit[i], Math.Max(lowerEstimationLimit[i], estimated[i]));
        }
        evaluator.Add(original, estimated);
      }

      double quality = evaluator.Value;
      return quality;
    }

    public static void CalculateScalingParameters(IEnumerable<double[]> originalValues, IEnumerable<double[]> estimatedValues, ref double[] beta, ref double[] alpha) {
      List<OnlineMeanAndVarianceCalculator> estimatedVarianceEvaluators = new List<OnlineMeanAndVarianceCalculator>();
      List<OnlineCovarianceEvaluator> covarianceEvaluators = new List<OnlineCovarianceEvaluator>();
      List<OnlineMeanAndVarianceCalculator> originalMeanCalculators = new List<OnlineMeanAndVarianceCalculator>();
      int[] cnt = null;

      var estimatedEnumerator = estimatedValues.GetEnumerator();
      var originalEnumerator = originalValues.GetEnumerator();
      while (estimatedEnumerator.MoveNext() & originalEnumerator.MoveNext()) {
        double[] original = originalEnumerator.Current;
        double[] estimated = estimatedEnumerator.Current;
        int dimension = original.Length;
        // initialize
        if (cnt == null) {
          cnt = new int[dimension];
          for (int i = 0; i < dimension; i++) {
            estimatedVarianceEvaluators.Add(new OnlineMeanAndVarianceCalculator());
            covarianceEvaluators.Add(new OnlineCovarianceEvaluator());
            originalMeanCalculators.Add(new OnlineMeanAndVarianceCalculator());
          }
        } else if (cnt.Length == dimension) {
          for (int component = 0; component < dimension; component++) {
            if (IsValidValue(original[component]) && IsValidValue(estimated[component])) {
              cnt[component]++;
              estimatedVarianceEvaluators[component].Add(estimated[component]);
              covarianceEvaluators[component].Add(original[component], estimated[component]);
              originalMeanCalculators[component].Add(original[component]);
            }
          }
        } else {
          throw new ArgumentException("Dimension of input array doesn't match");
        }
      }

      if (estimatedEnumerator.MoveNext() || originalEnumerator.MoveNext())
        throw new InvalidOperationException("Number of elements in estimated and original series doesn't match.");
      if (cnt == null) throw new ArgumentException("No elements in estimated and original.");

      for (int component = 0; component < cnt.Length; component++) {
        if (cnt[component] < 2) {
          alpha[component] = 0;
          beta[component] = 1;
        } else {
          if (estimatedVarianceEvaluators[component].PopulationVariance.IsAlmost(0.0))
            beta[component] = 1;
          else
            beta[component] = covarianceEvaluators[component].Covariance / estimatedVarianceEvaluators[component].PopulationVariance;

          alpha[component] = originalMeanCalculators[component].Mean - beta[component] * estimatedVarianceEvaluators[component].Mean;
        }
      }
    }

    private static bool IsValidValue(double d) {
      return !double.IsInfinity(d) && !double.IsNaN(d) && d > -1.0E07 && d < 1.0E07;  // don't consider very large or very small values for scaling
    }
  }
}
